# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import io
import os
import subprocess

from twisted.internet import defer
from twisted.internet import error
from twisted.internet import protocol
from twisted.python import log
from twisted.python import runtime

from buildbot.util import unicode2bytes


class RunProcessPP(protocol.ProcessProtocol):
    def __init__(self, run_process, initial_stdin=None):
        self.run_process = run_process
        self.initial_stdin = initial_stdin

    def connectionMade(self):
        if self.initial_stdin:
            self.transport.write(self.initial_stdin)
        self.transport.closeStdin()

    def outReceived(self, data):
        self.run_process.add_stdout(data)

    def errReceived(self, data):
        self.run_process.add_stderr(data)

    def processEnded(self, reason):
        self.run_process.process_ended(reason.value.signal, reason.value.exitCode)


class RunProcess:
    TIMEOUT_KILL = 5
    interrupt_signal = "KILL"

    def __init__(
        self,
        reactor,
        command,
        workdir=None,
        env=None,
        collect_stdout=True,
        collect_stderr=True,
        stderr_is_error=False,
        io_timeout=300,
        runtime_timeout=3600,
        sigterm_timeout=5,
        initial_stdin=None,
        use_pty=False,
    ):
        self._reactor = reactor
        self.command = command

        self.workdir = workdir
        self.process = None

        self.environ = env

        self.initial_stdin = initial_stdin

        self.output_stdout = None
        self.consumer_stdout = None

        if collect_stdout is True:
            self.output_stdout = io.BytesIO()
            self.consumer_stdout = self.output_stdout.write
        elif callable(collect_stdout):
            self.consumer_stdout = collect_stdout

        self.output_stderr = None
        self.consumer_stderr = None

        if collect_stderr is True:
            self.output_stderr = io.BytesIO()
            self.consumer_stderr = self.output_stderr.write
        elif callable(collect_stderr):
            self.consumer_stderr = collect_stderr

        self.stderr_is_error = stderr_is_error

        self.io_timeout = io_timeout
        self.io_timer = None

        self.sigterm_timeout = sigterm_timeout
        self.sigterm_timer = None

        self.runtime_timeout = runtime_timeout
        self.runtime_timer = None

        self.killed = False
        self.kill_timer = None
        self.use_pty = use_pty

        self.result_signal = None
        self.result_rc = None

    def __repr__(self):
        return f"<{self.__class__.__name__} '{self.command}'>"

    def get_os_env(self):
        return os.environ

    def resolve_environment(self, env):
        os_env = self.get_os_env()
        if env is None:
            return os_env.copy()

        new_env = {}
        for key, value in os_env.items():
            if key not in env or env[key] is not None:
                new_env[key] = value
        for key, value in env.items():
            if value is not None:
                new_env[key] = value
        return new_env

    def start(self):
        self.deferred = defer.Deferred()
        try:
            self._start_command()
        except Exception as e:
            self.deferred.errback(e)
        return self.deferred

    def _start_command(self):
        self.pp = RunProcessPP(self, initial_stdin=self.initial_stdin)

        environ = self.resolve_environment(self.environ)

        # $PWD usually indicates the current directory; spawnProcess may not
        # update this value, though, so we set it explicitly here.  This causes
        # weird problems (bug #456) on msys
        if not environ.get('MACHTYPE', None) == 'i686-pc-msys' and self.workdir is not None:
            environ['PWD'] = os.path.abspath(self.workdir)

        argv = unicode2bytes(self.command)
        self.process = self._reactor.spawnProcess(
            self.pp, argv[0], argv, environ, self.workdir, usePTY=self.use_pty
        )

        if self.io_timeout:
            self.io_timer = self._reactor.callLater(self.io_timeout, self.io_timed_out)

        if self.runtime_timeout:
            self.runtime_timer = self._reactor.callLater(
                self.runtime_timeout, self.runtime_timed_out
            )

    def add_stdout(self, data):
        if self.consumer_stdout is not None:
            self.consumer_stdout(data)

        if self.io_timer:
            self.io_timer.reset(self.io_timeout)

    def add_stderr(self, data):
        if self.consumer_stderr is not None:
            self.consumer_stderr(data)

        if self.stderr_is_error:
            self.kill('command produced stderr which is interpreted as error')

        if self.io_timer:
            self.io_timer.reset(self.io_timeout)

    def _build_result(self, rc):
        if self.output_stdout is not None and self.output_stderr is not None:
            return (rc, self.output_stdout.getvalue(), self.output_stderr.getvalue())
        if self.output_stdout is not None:
            return (rc, self.output_stdout.getvalue())
        if self.output_stderr is not None:
            return (rc, self.output_stderr.getvalue())
        return rc

    def process_ended(self, sig, rc):
        self.result_signal = sig
        self.result_rc = rc

        if self.killed and rc == 0:
            log.msg("process was killed, but exited with status 0; faking a failure")

            # windows returns '1' even for signalled failures, while POSIX returns -1
            if runtime.platformType == 'win32':
                rc = 1
            else:
                rc = -1

        if sig is not None:
            rc = -1

        self._cancel_timers()
        d = self.deferred
        self.deferred = None
        if d:
            d.callback(self._build_result(rc))
        else:
            log.err(f"{self}: command finished twice")

    def failed(self, why):
        self._cancel_timers()
        d = self.deferred
        self.deferred = None
        if d:
            d.errback(why)
        else:
            log.err(f"{self}: command finished twice")

    def io_timed_out(self):
        self.io_timer = None
        msg = f"{self}: command timed out: {self.io_timeout} seconds without output"
        self.kill(msg)

    def runtime_timed_out(self):
        self.runtime_timer = None
        msg = f"{self}: command timed out: {self.runtime_timeout} seconds elapsed"
        self.kill(msg)

    def is_dead(self):
        if self.process.pid is None:
            return True
        pid = int(self.process.pid)
        try:
            os.kill(pid, 0)
        except OSError:
            return True
        return False

    def check_process_was_killed(self):
        self.sigterm_timer = None
        if not self.is_dead():
            if not self.send_signal(self.interrupt_signal):
                log.msg(f"{self}: failed to kill process again")

        self.cleanup_killed_process()

    def cleanup_killed_process(self):
        if runtime.platformType == "posix":
            # we only do this under posix because the win32eventreactor
            # blocks here until the process has terminated, while closing
            # stderr. This is weird.
            self.pp.transport.loseConnection()

        if self.deferred:
            # finished ought to be called momentarily. Just in case it doesn't,
            # set a timer which will abandon the command.
            self.kill_timer = self._reactor.callLater(self.TIMEOUT_KILL, self.kill_timed_out)

    def send_signal(self, interrupt_signal):
        success = False

        log.msg(f'{self}: killing process using {interrupt_signal}')

        if runtime.platformType == "win32":
            pid = self.process.pid
            if interrupt_signal is not None and pid is not None:
                try:
                    if interrupt_signal == "TERM":
                        # TODO: blocks
                        subprocess.check_call(f"TASKKILL /PID {pid} /T")
                        success = True
                    elif interrupt_signal == "KILL":
                        # TODO: blocks
                        subprocess.check_call(f"TASKKILL /F /PID {pid} /T")
                        success = True
                except subprocess.CalledProcessError as e:
                    # taskkill may return 128 or 255 as exit code when the child has already exited.
                    # We can't handle this race condition in any other way than just interpreting
                    # the kill action as successful
                    if e.returncode in (128, 255):
                        log.msg(f"{self} taskkill didn't find pid {pid} to kill")
                        success = True
                    else:
                        raise

        # try signalling the process itself (works on Windows too, sorta)
        if not success:
            try:
                self.process.signalProcess(interrupt_signal)
                success = True
            except OSError as e:
                log.err(f"{self}: from process.signalProcess: {e}")
                # could be no-such-process, because they finished very recently
            except error.ProcessExitedAlready:
                log.msg(f"{self}: process exited already - can't kill")

                # the process has already exited, and likely finished() has
                # been called already or will be called shortly

        return success

    def kill(self, msg):
        log.msg(f'{self}: killing process because {msg}')
        self._cancel_timers()

        self.killed = True

        if self.sigterm_timeout is not None:
            self.send_signal("TERM")
            self.sigterm_timer = self._reactor.callLater(
                self.sigterm_timeout, self.check_process_was_killed
            )
        else:
            if not self.send_signal(self.interrupt_signal):
                log.msg(f"{self}: failed to kill process")

            self.cleanup_killed_process()

    def kill_timed_out(self):
        self.kill_timer = None
        log.msg(f"{self}: attempted to kill process, but it wouldn't die")

        self.failed(RuntimeError(f"SIG{self.interrupt_signal} failed to kill process"))

    def _cancel_timers(self):
        for name in ('io_timer', 'kill_timer', 'runtime_timer', 'sigterm_timer'):
            timer = getattr(self, name, None)
            if timer:
                timer.cancel()
                setattr(self, name, None)


def create_process(*args, **kwargs):
    return RunProcess(*args, **kwargs)


def run_process(*args, **kwargs):
    process = create_process(*args, **kwargs)
    return process.start()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

"""add codebases, codebase_commits and codebase_branches tables

Revision ID: 067
Revises: 066

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "067"
down_revision = "066"
branch_labels = None
depends_on = None


def upgrade() -> None:
    hash_length = 40

    op.create_table(
        'codebases',
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column(
            'projectid',
            sa.Integer,
            sa.ForeignKey('projects.id', ondelete='CASCADE'),
            nullable=False,
        ),
        sa.Column('name', sa.Text, nullable=False),
        sa.Column('slug', sa.String(50), nullable=False),
        sa.Column('name_hash', sa.String(hash_length), nullable=False),
        mysql_DEFAULT_CHARSET='utf8',
    )

    op.create_table(
        'codebase_commits',
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column(
            'codebaseid',
            sa.Integer,
            sa.ForeignKey('codebases.id', ondelete='CASCADE'),
            nullable=False,
        ),
        sa.Column('author', sa.String(255), nullable=False),
        sa.Column('committer', sa.String(255), nullable=True),
        sa.Column('comments', sa.Text, nullable=False),
        sa.Column('when_timestamp', sa.Integer, nullable=False),
        sa.Column('revision', sa.String(70), nullable=False),
        sa.Column(
            'parent_commitid',
            sa.Integer,
            sa.ForeignKey('codebase_commits.id', ondelete='SET NULL'),
            nullable=True,
        ),
        mysql_DEFAULT_CHARSET='utf8',
    )

    op.create_table(
        'codebase_branches',
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column(
            'codebaseid',
            sa.Integer,
            sa.ForeignKey('codebases.id', ondelete='CASCADE'),
            nullable=False,
        ),
        sa.Column('name', sa.String(255), nullable=False),
        sa.Column('name_hash', sa.String(hash_length), nullable=False),
        sa.Column(
            'commitid',
            sa.Integer,
            sa.ForeignKey('codebase_commits.id', ondelete='SET NULL'),
            nullable=True,
        ),
        sa.Column('last_timestamp', sa.Integer, nullable=False),
        mysql_DEFAULT_CHARSET='utf8',
    )

    op.create_index(
        'codebases_projects_name_hash',
        'codebases',
        ['projectid', 'name_hash'],
        unique=True,
    )

    op.create_index(
        'codebase_commits_unique',
        'codebase_commits',
        ['codebaseid', 'revision'],
        unique=True,
    )

    op.create_index(
        'codebase_branches_unique',
        'codebase_branches',
        ['codebaseid', 'name_hash'],
        unique=True,
    )


def downgrade() -> None:
    op.drop_table('codebases')
    op.drop_table('codebase_commits')
    op.drop_table('codebase_branches')
    op.drop_index('codebases_projects_name_hash')
    op.drop_index('codebase_commits_unique')
    op.drop_index('codebase_branches_unique')



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from dataclasses import field

import sqlalchemy as sa
from twisted.internet import defer

from buildbot.db import base
from buildbot.util.sautils import hash_columns
from buildbot.warnings import warn_deprecated


@dataclass
class BuilderModel:
    id: int
    name: str
    description: str | None = None
    description_format: str | None = None
    description_html: str | None = None
    projectid: int | None = None
    tags: list[str] = field(default_factory=list)
    masterids: list[int] = field(default_factory=list)

    # For backward compatibility
    def __getitem__(self, key: str):
        warn_deprecated(
            '4.1.0',
            (
                'BuildersConnectorComponent getBuilder and getBuilders '
                'no longer return Builder as dictionnaries. '
                'Usage of [] accessor is deprecated: please access the member directly'
            ),
        )

        if hasattr(self, key):
            return getattr(self, key)

        raise KeyError(key)


class BuildersConnectorComponent(base.DBConnectorComponent):
    def findBuilderId(self, name, autoCreate=True):
        tbl = self.db.model.builders
        name_hash = hash_columns(name)
        return self.findSomethingId(
            tbl=tbl,
            whereclause=(tbl.c.name_hash == name_hash),
            insert_values={"name": name, "name_hash": name_hash},
            autoCreate=autoCreate,
        )

    @defer.inlineCallbacks
    def updateBuilderInfo(
        self, builderid, description, description_format, description_html, projectid, tags
    ):
        # convert to tag IDs first, as necessary
        def toTagid(tag):
            if isinstance(tag, int):
                return defer.succeed(tag)
            ssConnector = self.master.db.tags
            return ssConnector.findTagId(tag)

        tagsids = [
            r[1]
            for r in (
                yield defer.DeferredList(
                    [toTagid(tag) for tag in tags], fireOnOneErrback=True, consumeErrors=True
                )
            )
        ]

        def thd(conn):
            builders_tbl = self.db.model.builders
            builders_tags_tbl = self.db.model.builders_tags
            transaction = conn.begin()

            q = builders_tbl.update().where(builders_tbl.c.id == builderid)
            conn.execute(
                q.values(
                    description=description,
                    description_format=description_format,
                    description_html=description_html,
                    projectid=projectid,
                )
            ).close()
            # remove previous builders_tags
            conn.execute(
                builders_tags_tbl.delete().where(builders_tags_tbl.c.builderid == builderid)
            ).close()

            # add tag ids
            if tagsids:
                conn.execute(
                    builders_tags_tbl.insert(),
                    [{"builderid": builderid, "tagid": tagid} for tagid in tagsids],
                ).close()

            transaction.commit()

        return (yield self.db.pool.do(thd))

    @defer.inlineCallbacks
    def getBuilder(self, builderid: int):
        bldrs: list[BuilderModel] = yield self.getBuilders(_builderid=builderid)
        if bldrs:
            return bldrs[0]
        return None

    # returns a Deferred that returns None
    def addBuilderMaster(self, builderid=None, masterid=None):
        def thd(conn, no_recurse=False):
            try:
                tbl = self.db.model.builder_masters
                q = tbl.insert()
                conn.execute(q.values(builderid=builderid, masterid=masterid))
                conn.commit()
            except (sa.exc.IntegrityError, sa.exc.ProgrammingError):
                conn.rollback()

        return self.db.pool.do(thd)

    # returns a Deferred that returns None
    def removeBuilderMaster(self, builderid=None, masterid=None):
        def thd(conn, no_recurse=False):
            tbl = self.db.model.builder_masters
            conn.execute(
                tbl.delete().where(tbl.c.builderid == builderid, tbl.c.masterid == masterid)
            )

        return self.db.pool.do_with_transaction(thd)

    def getBuilders(
        self,
        masterid: int | None = None,
        projectid: int | None = None,
        workerid: int | None = None,
        _builderid: int | None = None,
    ) -> defer.Deferred[list[BuilderModel]]:
        def thd(conn) -> list[BuilderModel]:
            bldr_tbl = self.db.model.builders
            bm_tbl = self.db.model.builder_masters
            builders_tags_tbl = self.db.model.builders_tags
            tags_tbl = self.db.model.tags
            configured_workers_tbl = self.db.model.configured_workers

            j = bldr_tbl.outerjoin(bm_tbl)
            # if we want to filter by masterid, we must join to builder_masters
            # again, so we can still get the full set of masters for each
            # builder
            if masterid is not None:
                limiting_bm_tbl = bm_tbl.alias('limiting_bm')
                j = j.join(limiting_bm_tbl, onclause=bldr_tbl.c.id == limiting_bm_tbl.c.builderid)
            if workerid is not None:
                j = j.join(configured_workers_tbl)
            q = (
                sa.select(
                    bldr_tbl.c.id,
                    bldr_tbl.c.name,
                    bldr_tbl.c.description,
                    bldr_tbl.c.description_format,
                    bldr_tbl.c.description_html,
                    bldr_tbl.c.projectid,
                    bm_tbl.c.masterid,
                )
                .select_from(j)
                .order_by(bldr_tbl.c.id, bm_tbl.c.masterid)
            )
            if masterid is not None:
                # filter the masterid from the limiting table
                q = q.where(limiting_bm_tbl.c.masterid == masterid)
            if projectid is not None:
                q = q.where(bldr_tbl.c.projectid == projectid)
            if workerid is not None:
                q = q.where(configured_workers_tbl.c.workerid == workerid)
            if _builderid is not None:
                q = q.where(bldr_tbl.c.id == _builderid)

            # build up a intermediate builder id -> tag names map (fixes performance issue #3396)
            bldr_id_to_tags = defaultdict(list)
            bldr_q = sa.select(builders_tags_tbl.c.builderid, tags_tbl.c.name)
            bldr_q = bldr_q.select_from(tags_tbl.join(builders_tags_tbl))

            for bldr_id, tag in conn.execute(bldr_q).fetchall():
                bldr_id_to_tags[bldr_id].append(tag)

            # now group those by builderid, aggregating by masterid
            rv: list[BuilderModel] = []
            last: BuilderModel | None = None
            for row in conn.execute(q).fetchall():
                if not last or row.id != last.id:
                    last = BuilderModel(
                        id=row.id,
                        name=row.name,
                        description=row.description,
                        description_format=row.description_format,
                        description_html=row.description_html,
                        projectid=row.projectid,
                        tags=bldr_id_to_tags[row.id],
                    )
                    rv.append(last)
                if row.masterid:
                    last.masterids.append(row.masterid)
            return rv

        return self.db.pool.do(thd)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import json
import os
import textwrap
from unittest import mock

from autobahn.wamp.exception import TransportLost
from autobahn.wamp.types import SubscribeOptions
from twisted.internet import defer
from twisted.trial import unittest

from buildbot.mq import wamp
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.wamp import connector


class FakeEventDetails:
    def __init__(self, topic):
        self.topic = topic


class ComparableSubscribeOptions(SubscribeOptions):
    def __eq__(self, other):
        if not isinstance(other, SubscribeOptions):
            return False
        return self.match == other.match

    __repr__ = SubscribeOptions.__str__


class FakeSubscription:
    def __init__(self):
        self.exception_on_unsubscribe = None

    def unsubscribe(self):
        if self.exception_on_unsubscribe is not None:
            raise self.exception_on_unsubscribe()


class TestException(Exception):
    pass


class FakeWampConnector:
    # a fake wamp connector with only one queue
    def __init__(self):
        self.subscriptions = []

    def topic_match(self, topic):
        topic = topic.split(".")
        owntopic = self.topic.split(".")
        if len(topic) != len(owntopic):
            return False
        for i, itopic in enumerate(topic):
            if owntopic[i] != "" and itopic != owntopic[i]:
                return False
        return True

    def subscribe(self, callback, topic=None, options=None):
        # record the topic, and to make sure subsequent publish
        # are correct
        self.topic = topic
        # we record the qref_cb
        self.qref_cb = callback

        subs = FakeSubscription()
        self.subscriptions.append(subs)
        return subs

    def publish(self, topic, data, options=None):
        # make sure the topic is compatible with what was subscribed
        assert self.topic_match(topic)
        self.last_data = data
        details = FakeEventDetails(topic=topic)
        self.qref_cb(json.loads(json.dumps(data)), details=details)


class TopicMatch(unittest.TestCase):
    # test unit tests

    def test_topic_match(self):
        matches = [
            ("a.b.c", "a.b.c"),
            ("a..c", "a.c.c"),
            ("a.b.", "a.b.c"),
            (".b.", "a.b.c"),
        ]
        for i, j in matches:
            w = FakeWampConnector()
            w.topic = i
            self.assertTrue(w.topic_match(j))

    def test_topic_not_match(self):
        matches = [
            ("a.b.c", "a.b.d"),
            ("a..c", "a.b.d"),
            ("a.b.", "a.c.c"),
            (".b.", "a.a.c"),
        ]
        for i, j in matches:
            w = FakeWampConnector()
            w.topic = i
            self.assertFalse(w.topic_match(j))


class WampMQ(TestReactorMixin, unittest.TestCase):
    """
    Stimulate the code with a fake wamp router:
    A router which only accepts one subscriber on one topic
    """

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self)
        self.master.wamp = FakeWampConnector()
        self.mq = wamp.WampMQ()
        yield self.mq.setServiceParent(self.master)
        yield self.mq.startService()

        @defer.inlineCallbacks
        def cleanup():
            if self.mq.running:
                yield self.mq.stopService()

        self.addCleanup(cleanup)

    @defer.inlineCallbacks
    def test_startConsuming_basic(self):
        self.master.wamp.subscribe = mock.Mock()
        yield self.mq.startConsuming(None, ('a', 'b'))
        options = ComparableSubscribeOptions(details_arg='details')
        self.master.wamp.subscribe.assert_called_with(
            mock.ANY, 'org.buildbot.mq.a.b', options=options
        )

    @defer.inlineCallbacks
    def test_startConsuming_wildcard(self):
        self.master.wamp.subscribe = mock.Mock()
        yield self.mq.startConsuming(None, ('a', None))
        options = ComparableSubscribeOptions(match="wildcard", details_arg='details')
        self.master.wamp.subscribe.assert_called_with(
            mock.ANY, 'org.buildbot.mq.a.', options=options
        )

    @defer.inlineCallbacks
    def test_forward_data(self):
        callback = mock.Mock()
        yield self.mq.startConsuming(callback, ('a', 'b'))
        # _produce returns a deferred
        yield self.mq._produce(('a', 'b'), 'foo')
        # calling produce should eventually call the callback with decoding of
        # topic
        callback.assert_called_with(('a', 'b'), 'foo')
        self.assertEqual(self.master.wamp.last_data, 'foo')

    @defer.inlineCallbacks
    def test_unsubscribe_ignores_transport_lost(self):
        callback = mock.Mock()
        consumer = yield self.mq.startConsuming(callback, ('a', 'b'))

        self.assertEqual(len(self.master.wamp.subscriptions), 1)
        self.master.wamp.subscriptions[0].exception_on_unsubscribe = TransportLost

        yield consumer.stopConsuming()

    @defer.inlineCallbacks
    def test_unsubscribe_logs_exceptions(self):
        callback = mock.Mock()
        consumer = yield self.mq.startConsuming(callback, ('a', 'b'))

        self.assertEqual(len(self.master.wamp.subscriptions), 1)
        self.master.wamp.subscriptions[0].exception_on_unsubscribe = TestException

        yield consumer.stopConsuming()

        self.assertEqual(len(self.flushLoggedErrors(TestException)), 1)

    @defer.inlineCallbacks
    def test_forward_data_wildcard(self):
        callback = mock.Mock()
        yield self.mq.startConsuming(callback, ('a', None))
        # _produce returns a deferred
        yield self.mq._produce(('a', 'b'), 'foo')
        # calling produce should eventually call the callback with decoding of
        # topic
        callback.assert_called_with(('a', 'b'), 'foo')
        self.assertEqual(self.master.wamp.last_data, 'foo')

    @defer.inlineCallbacks
    def test_waits_for_called_callback(self):
        def callback(_, __):
            return defer.succeed(None)

        yield self.mq.startConsuming(callback, ('a', None))
        yield self.mq._produce(('a', 'b'), 'foo')
        self.assertEqual(self.master.wamp.last_data, 'foo')

        d = self.mq.stopService()
        self.assertTrue(d.called)

    @defer.inlineCallbacks
    def test_waits_for_non_called_callback(self):
        d1 = defer.Deferred()

        def callback(_, __):
            return d1

        yield self.mq.startConsuming(callback, ('a', None))
        yield self.mq._produce(('a', 'b'), 'foo')
        self.assertEqual(self.master.wamp.last_data, 'foo')

        d = self.mq.stopService()
        self.assertFalse(d.called)
        d1.callback(None)
        self.assertTrue(d.called)


class FakeConfig:
    mq = {"type": 'wamp', "router_url": 'wss://foo', "realm": 'realm1'}


class WampMQReal(TestReactorMixin, unittest.TestCase):
    """
    Tests a little bit more painful to run, but which involve real communication with
    a wamp router
    """

    HOW_TO_RUN = textwrap.dedent("""\
        define WAMP_ROUTER_URL to a wamp router to run this test
        > crossbar init
        > crossbar start &
        > export WAMP_ROUTER_URL=ws://localhost:8080/ws
        > trial buildbot.unit.test_mq_wamp""")

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        if "WAMP_ROUTER_URL" not in os.environ:
            raise unittest.SkipTest(self.HOW_TO_RUN)
        self.master = yield fakemaster.make_master(self)
        self.mq = wamp.WampMQ()
        yield self.mq.setServiceParent(self.master)
        self.connector = self.master.wamp = connector.WampConnector()
        yield self.connector.setServiceParent(self.master)

        yield self.master.startService()
        self.addCleanup(self.master.stopService)

        config = FakeConfig()
        config.mq['router_url'] = os.environ["WAMP_ROUTER_URL"]
        yield self.connector.reconfigServiceWithBuildbotConfig(config)

    @defer.inlineCallbacks
    def test_forward_data(self):
        d = defer.Deferred()
        callback = mock.Mock(side_effect=lambda *a, **kw: d.callback(None))
        yield self.mq.startConsuming(callback, ('a', 'b'))
        # _produce returns a deferred
        yield self.mq._produce(('a', 'b'), 'foo')
        # calling produce should eventually call the callback with decoding of
        # topic
        yield d
        callback.assert_called_with(('a', 'b'), 'foo')

    @defer.inlineCallbacks
    def test_forward_data_wildcard(self):
        d = defer.Deferred()
        callback = mock.Mock(side_effect=lambda *a, **kw: d.callback(None))
        yield self.mq.startConsuming(callback, ('a', None))
        # _produce returns a deferred
        yield self.mq._produce(('a', 'b'), 'foo')
        # calling produce should eventually call the callback with decoding of
        # topic
        yield d
        callback.assert_called_with(('a', 'b'), 'foo')
